"""Test cases for the decimal filter."""
# pylint: disable=missing-class-docstring,missing-function-docstring
import unittest

from babel import UnknownLocaleError

from liquid import Environment

from liquid_babel.filters import Number
from liquid_babel.filters import number


class NumberFilterTestCase(unittest.TestCase):
    def test_defaults(self) -> None:
        """Test that the default locale and format use en_US."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ '374881.01' | decimal }}")
        result = template.render()
        self.assertEqual(result, "374,881.01")

    def test_set_default_locale(self) -> None:
        """Test that we can set the default locale."""
        env = Environment()
        env.add_filter("decimal", Number(default_locale="de"))
        template = env.from_string("{{ '374881.01' | decimal }}")
        result = template.render()
        self.assertEqual(result, "374.881,01")

    def test_get_default_locale_from_context(self) -> None:
        """Test that we can get a default locale from context."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ '374881.01' | decimal }}")
        result = template.render(locale="de")
        self.assertEqual(result, "374.881,01")

    def test_default_decimal_quantization(self) -> None:
        """Test that decimal quantization defaults to False."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ '2.2346' | decimal }}")
        result = template.render()
        self.assertEqual(result, "2.2346")

    def test_set_default_decimal_quantization(self) -> None:
        """Test that we can set the default decimal quantization."""
        env = Environment()
        env.add_filter("decimal", Number(default_decimal_quantization=True))
        template = env.from_string("{{ '2.2346' | decimal }}")
        result = template.render()
        self.assertEqual(result, "2.235")

    def test_get_decimal_quantization_from_context(self) -> None:
        """Test that we can get decimal quantization from context."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ '2.2346' | decimal }}")
        result = template.render(decimal_quantization=True)
        self.assertEqual(result, "2.235")

    def test_parse_string(self) -> None:
        """Test parse a string to a decimal with the default input locale."""
        env = Environment()
        env.add_filter("decimal", number)
        # Parse as en_US
        template = env.from_string("{{ '10,000.00' | decimal }}")
        result = template.render(locale="de")
        # Render as de
        self.assertEqual(result, "10.000")

    def test_parse_string_with_input_locale_from_context(self) -> None:
        """Test parse a string to a decimal with an input locale from context."""
        env = Environment()
        env.add_filter("decimal", number)
        # Parse as de
        template = env.from_string("{{ '10.000,00' | decimal }}")
        result = template.render(locale="en_US", input_locale="de")
        # Render as en_US
        self.assertEqual(result, "10,000")

    def test_set_default_input_locale(self) -> None:
        """Test that we can set the default input locale."""
        env = Environment()
        # Parse as de
        env.add_filter("decimal", Number(default_input_locale="de"))
        template = env.from_string("{{ '10.000,00' | decimal }}")
        result = template.render(locale="en_US")
        # Render as en_US
        self.assertEqual(result, "10,000")

    def test_unknown_locale(self) -> None:
        """Test that an unknown locale raises an exception early."""
        env = Environment()
        with self.assertRaises(UnknownLocaleError):
            env.add_filter("decimal", Number(default_locale="nosuchthing"))

    def test_unknown_locale_from_context(self) -> None:
        """Test that an unknown locale falls back to default."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ '1.99' | decimal }}")
        result = template.render(locale="nosuchthing")
        self.assertEqual(result, "1.99")

    def test_invalid_string_left_value(self) -> None:
        """Test that invalid decimals default to zero"""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ 'not a number' | decimal }}")
        result = template.render()
        self.assertEqual(result, "0")

    def test_arbitrary_object_left_value(self) -> None:
        """Test that arbitrary objects default to zero."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ obj | decimal }}")
        result = template.render(obj=object())
        self.assertEqual(result, "0")

    def test_float_left_value(self) -> None:
        """Test that float left values are parsed."""
        env = Environment()
        env.add_filter("decimal", number)
        template = env.from_string("{{ 1.5 | decimal }}")
        result = template.render()
        self.assertEqual(result, "1.5")



"""A currency formatting filter for Python Liquid."""
from decimal import Decimal

from typing import cast
from typing import Optional
from typing import Union

from babel import numbers
from babel import Locale
from babel import UnknownLocaleError

from liquid import Context
from liquid.context import is_undefined

from liquid.filter import liquid_filter
from liquid.filter import num_arg
from liquid.filter import with_context


# pylint: disable=too-few-public-methods too-many-instance-attributes
@with_context
class Currency:
    """A Liquid filter for formatting currency values.

    :param currency_code_var: The name of a render context variable that resolves
        to the current currency code. Defaults to ``"currency_code"``.
    :type currency_code_var: str
    :param default_currency_code: A fallback currency code if ``currency_code_var``
        can not be resolved. Defaults to ``"USD"``.
    :type default_currency_code: str
    :param locale_var: The name of a render context variable that resolves to the
        current locale. Defaults to ``"locale"``.
    :type locale_var: str
    :param default_locale : A fallback locale to use if ``locale_var`` can not be
        resolved. Defaults to ``"en_US"``.
    :type default_locale: str
    :param format_var: The name of a render context variable that resolves to the
        current currency format string. Defaults to ``"currency_format"``.
    :type format_var: str
    :param default_format: A fallback currency format that is used if ``format_var``
        can not be resolved. Defaults to ``None``, which means the standard format for
        the current locale will be used.
    :type default_format: str | None
    :param currency_digits: Indicates if the format should override locale specific
        trailing digit behavior. Defaults to ``False``.
    :type currency_digits: bool
    :param input_locale_var: The name of a render context variable that resolves to
        a locale suitable for parsing input strings to decimals. Defaults to
        ``"input_locale"``.
    :type input_locale_var: str
    :param default_input_locale: A fallback locale to use if ``input_locale_var``
        can not be resolved. Defaults to ``"en_US"``.
    """

    def __init__(
        self,
        *,
        currency_code_var: str = "currency_code",
        default_currency_code: str = "USD",
        locale_var: str = "locale",
        default_locale: str = "en_US",
        format_var: str = "currency_format",
        default_format: Optional[str] = None,
        currency_digits: bool = True,
        input_locale_var: str = "input_locale",
        default_input_locale: str = "en_US",
    ) -> None:
        self.currency_code_var = currency_code_var
        self.default_currency_code = default_currency_code
        self.locale_var = locale_var
        self.default_locale = Locale.parse(default_locale)
        self.format_var = format_var
        self.default_format = default_format
        self.currency_digits = currency_digits
        self.input_locale_var = input_locale_var
        self.default_input_locale = Locale.parse(default_input_locale)

    @liquid_filter
    def __call__(
        self,
        left: object,
        *,
        context: Context,
        group_separator: bool = True,
    ) -> str:
        locale = self._resolve_locale(
            context,
            self.locale_var,
            default=self.default_locale,
        )
        input_locale = self._resolve_locale(
            context,
            self.input_locale_var,
            default=self.default_input_locale,
        )
        _format = context.resolve(self.format_var, default=self.default_format)
        currency_code = context.resolve(
            self.currency_code_var,
            default=self.default_currency_code,
        )

        return cast(
            str,
            numbers.format_currency(
                _parse_decimal(left, input_locale),
                currency_code,
                format=_format,
                locale=locale,
                group_separator=group_separator,
                currency_digits=self.currency_digits,
            ),
        )

    def _resolve_locale(
        self,
        context: Context,
        locale_var: str,
        default: Locale,
    ) -> Locale:
        _locale = context.resolve(locale_var)
        if not is_undefined(_locale):
            try:
                locale = Locale.parse(_locale)
            except UnknownLocaleError:
                locale = default
        else:
            locale = default

        return cast(Locale, locale)


def _parse_decimal(val: object, locale: Union[str, Locale]) -> Decimal:
    if isinstance(val, str):
        try:
            return cast(Decimal, numbers.parse_decimal(val, locale))
        except numbers.NumberFormatError:
            return Decimal(0)

    if isinstance(val, (Decimal, float, int)):
        return Decimal(val)

    # Give objects that implement __int__ etc. a chance.
    return Decimal(num_arg(val, 0))



# flake8: noqa
# pylint: disable=useless-import-alias,missing-module-docstring
from .currency import Currency as Currency
from .number import Number as Number

# For convenience. Use defaults.
currency = Currency()
number = Number()

# For convenience. Something akin to Shopify's money filters.
money = currency
money_with_currency = Currency(default_format="¤#,##0.00 ¤¤")
money_without_currency = Currency(default_format="#,##0.00")
money_without_trailing_zeros = Currency(
    default_format="¤#,###",
    currency_digits=False,
)



"""Test cases for the currency filter."""
# pylint: disable=missing-class-docstring,missing-function-docstring
import unittest

from babel import UnknownLocaleError

from liquid import Environment

from liquid_babel.filters import Currency
from liquid_babel.filters import currency
from liquid_babel.filters import money
from liquid_babel.filters import money_with_currency
from liquid_babel.filters import money_without_currency
from liquid_babel.filters import money_without_trailing_zeros


# pylint: disable=too-many-public-methods
class CurrencyFilterTestCase(unittest.TestCase):
    def test_default_currency_code_and_locale(self) -> None:
        """Test that the default currency code is USD and locale is en_US."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render()
        self.assertEqual(result, "$1.99")

    def test_set_currency_code_default(self) -> None:
        """Test that we can set the default currency code."""
        env = Environment()
        env.add_filter("currency", Currency(default_currency_code="GBP"))
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render()
        self.assertEqual(result, "£1.99")

    def test_currency_code_from_context(self) -> None:
        """Test that we can get a currency code from context."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render(currency_code="GBP")
        self.assertEqual(result, "£1.99")

    def test_unknown_currency_code(self) -> None:
        """Test that an unknown currency code is treated is prepended."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render(currency_code="nosuchthing")
        self.assertEqual(result, "nosuchthing1.99")

    def test_set_locale_default(self) -> None:
        """Test that we can set the default locale."""
        env = Environment()
        env.add_filter("currency", Currency(default_locale="de"))
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render()
        self.assertEqual(result, "1,99\xa0$")

    def test_locale_from_context(self) -> None:
        """Test that we can get a locale from context."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render(locale="de")
        self.assertEqual(result, "1,99\xa0$")

    def test_unknown_locale(self) -> None:
        """Test that an unknown locale raises an exception early."""
        env = Environment()
        with self.assertRaises(UnknownLocaleError):
            env.add_filter("currency", Currency(default_locale="nosuchthing"))

    def test_unknown_locale_from_context(self) -> None:
        """Test that an unknown locale falls back to default."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render(locale="nosuchthing")
        self.assertEqual(result, "$1.99")

    def test_set_format_default(self) -> None:
        """Test that we can set the default currency format."""
        env = Environment()
        env.add_filter("currency", Currency(default_format="¤¤ #,##0.00"))
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render()
        self.assertEqual(result, "USD 1.99")

    def test_format_from_context(self) -> None:
        """Test that we can get a currency format from context."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render(currency_format="¤¤ #,##0.00")
        self.assertEqual(result, "USD 1.99")

    def test_garbage_format(self) -> None:
        """Test that we don't fall over when given a garbage currency format."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 1.99 | currency }}")
        result = template.render(currency_format="bad format")
        self.assertEqual(result, "bad format1.99")

    def test_valid_string_left_value(self) -> None:
        """Test that we parse strings as decimals."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ '1.99' | currency }}")
        result = template.render()
        self.assertEqual(result, "$1.99")

    def test_invalid_string_left_value(self) -> None:
        """Test that invalid decimals default to zero"""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 'not a number' | currency }}")
        result = template.render()
        self.assertEqual(result, "$0.00")

    def test_arbitrary_object_left_value(self) -> None:
        """Test that arbitrary objects default to zero."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ obj | currency }}")
        result = template.render(obj=object())
        self.assertEqual(result, "$0.00")

    def test_group_separator_default(self) -> None:
        """Test that the group separator argument defaults to True."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 811375 | currency }}")
        result = template.render()
        self.assertEqual(result, "$811,375.00")

    def test_group_separator_argument(self) -> None:
        """Test that we can control group separators with a filter keyword argument."""
        env = Environment()
        env.add_filter("currency", currency)
        template = env.from_string("{{ 811375 | currency: group_separator: false }}")
        result = template.render()
        self.assertEqual(result, "$811375.00")

    def test_money(self) -> None:
        """Test that the convenience money filter formats similarly to shopify's."""
        env = Environment()
        env.add_filter("money", money)
        template = env.from_string("{{ 10 | money }}")
        result = template.render()
        self.assertEqual(result, "$10.00")

    def test_money_with_currency(self) -> None:
        """Test that the convenience money_with_currency filter formats similarly to shopify's."""
        env = Environment()
        env.add_filter("money_with_currency", money_with_currency)
        template = env.from_string("{{ 10 | money_with_currency }}")
        result = template.render(currency_code="CAD", locale="en_CA")
        self.assertEqual(result, "$10.00 CAD")

    def test_money_without_currency(self) -> None:
        """Test that the convenience money_without_currency filter formats similarly
        to shopify's."""
        env = Environment()
        env.add_filter("money_without_currency", money_without_currency)
        template = env.from_string("{{ 10 | money_without_currency }}")
        result = template.render()
        self.assertEqual(result, "10.00")

    def test_money_without_trailing_zeros(self) -> None:
        """Test that the convenience money_without_trailing_zeros filter formats similarly
        to shopify's."""
        env = Environment()
        env.add_filter("money_without_trailing_zeros", money_without_trailing_zeros)
        template = env.from_string("{{ 10 | money_without_trailing_zeros }}")
        result = template.render()
        self.assertEqual(result, "$10")

    def test_parse_string(self) -> None:
        """Test parse a string to a decimal with the default input locale."""
        env = Environment()
        env.add_filter("currency", currency)
        # Parse as en_US
        template = env.from_string("{{ '10,000.00' | currency }}")
        result = template.render(locale="de")
        # Render as de
        self.assertEqual(result, "10.000,00\xa0$")

    def test_parse_string_with_input_locale_from_context(self) -> None:
        """Test parse a string to a decimal with an input locale from context."""
        env = Environment()
        env.add_filter("currency", currency)
        # Parse as de
        template = env.from_string("{{ '10.000,00' | currency }}")
        result = template.render(locale="en_US", input_locale="de")
        # Render as en_US
        self.assertEqual(result, "$10,000.00")

    def test_set_default_input_locale(self) -> None:
        """Test that we can set the default input locale."""
        env = Environment()
        # Parse as de
        env.add_filter("currency", Currency(default_input_locale="de"))
        template = env.from_string("{{ '10.000,00' | currency }}")
        result = template.render(locale="en_US")
        # Render as en_US
        self.assertEqual(result, "$10,000.00")



"""A decimal formatting filter for Python Liquid."""
from decimal import Decimal

from typing import cast
from typing import Optional
from typing import Union

from babel import numbers
from babel import Locale
from babel import UnknownLocaleError

from liquid import Context
from liquid.context import is_undefined

from liquid.filter import liquid_filter
from liquid.filter import num_arg
from liquid.filter import with_context


# pylint: disable=too-few-public-methods too-many-instance-attributes
@with_context
class Number:
    """A Liquid filter for formatting decimal values.

    :param decimal_quantization_var: The name of a render context variable that
        resolves to the decimal quantization to be used. Defaults to
        ``"decimal_quantization"``.
    :type decimal_quantization_var: str
    :param default_decimal_quantization: A fallback decimal quantization if
        ``decimal_quantization_var`` can not be resolved. Defaults to ``False``.
    :type default_decimal_quantization: bool
    :param locale_var: The name of a render context variable that resolves to the
        current locale. Defaults to ``"locale"``.
    :type locale_var: str
    :param default_locale : A fallback locale to use if ``locale_var`` can not be
        resolved. Defaults to ``"en_US"``.
    :type default_locale: str
    :param format_var: The name of a render context variable that resolves to the
        current decimal format string. Defaults to ``"decimal_format"``.
    :type format_var: str
    :param default_format: A fallback decimal format that is used if ``format_var``
        can not be resolved. Defaults to ``None``, which means the standard format for
        the current locale will be used.
    :type default_format: str | None
    :param input_locale_var: The name of a render context variable that resolves to
        a locale suitable for parsing input strings to decimals. Defaults to
        ``"input_locale"``.
    :type input_locale_var: str
    :param default_input_locale: A fallback locale to use if ``input_locale_var``
        can not be resolved. Defaults to ``"en_US"``.
    """

    def __init__(
        self,
        *,
        decimal_quantization_var: str = "decimal_quantization",
        default_decimal_quantization: bool = False,
        locale_var: str = "locale",
        default_locale: str = "en_US",
        format_var: str = "decimal_format",
        default_format: Optional[str] = None,
        input_locale_var: str = "input_locale",
        default_input_locale: str = "en_US",
    ) -> None:
        self.decimal_quantization_var = decimal_quantization_var
        self.default_decimal_quantization = default_decimal_quantization
        self.locale_var = locale_var
        self.default_locale = Locale.parse(default_locale)
        self.format_var = format_var
        self.default_format = default_format
        self.input_locale_var = input_locale_var
        self.default_input_locale = Locale.parse(default_input_locale)

    @liquid_filter
    def __call__(
        self,
        left: object,
        *,
        context: Context,
        group_separator: bool = True,
    ) -> str:
        decimal_quantization = context.resolve(
            self.decimal_quantization_var,
            default=self.default_decimal_quantization,
        )
        locale = self._resolve_locale(
            context,
            self.locale_var,
            default=self.default_locale,
        )
        input_locale = self._resolve_locale(
            context,
            self.input_locale_var,
            default=self.default_input_locale,
        )
        _format = context.resolve(self.format_var, default=self.default_format)

        return cast(
            str,
            numbers.format_decimal(
                _parse_decimal(left, input_locale),
                format=_format,
                locale=locale,
                group_separator=group_separator,
                decimal_quantization=decimal_quantization,
            ),
        )

    def _resolve_locale(
        self,
        context: Context,
        locale_var: str,
        default: Locale,
    ) -> Locale:
        _locale = context.resolve(locale_var)
        if not is_undefined(_locale):
            try:
                locale = Locale.parse(_locale)
            except UnknownLocaleError:
                locale = default
        else:
            locale = default

        return cast(Locale, locale)


def _parse_decimal(val: object, locale: Union[str, Locale]) -> Decimal:
    if isinstance(val, str):
        try:
            return cast(Decimal, numbers.parse_decimal(val, locale))
        except numbers.NumberFormatError:
            return Decimal(0)

    if isinstance(val, (Decimal, float, int)):
        return Decimal(val)

    # Give objects that implement __int__ etc. a chance.
    return Decimal(num_arg(val, 0))
